
/*
 * emit.c
 *
 * (C) Copyright 1995 Archie L. Cobbs
 */

#include <stdio.h>
#include <assert.h>
#include "util.h"
#include "error.h"
#include "names.h"
#include "obj.h"
#include "asm.h"

/*
 * INTERNAL FUNCTIONS
 */

	static	void		EmitRel(int ins, IValue ivalue);

/*
 * EmitByte()
 *
 * Emit a byte into the current segment
 */

void
Emit(Byte byte)
{
	static int	gaveBssError = FALSE;

	if (gDebug)
		fprintf(stderr, "Emit: [%s]:$%04X: $%02X\n",
			gSegment == CodeSegment ? "CODE" :
			gSegment == DataSegment ? "DATA" :
			"BSS ", gAddress[gSegment], byte);

	switch (gSegment)
	{
		case CodeSegment:
		case DataSegment:
			if (gPass == 2)
				gSegBuffer[gSegment][gAddress[gSegment]] = byte;
			gAddress[gSegment]++;
			gaveBssError = FALSE;
			break;
		case BssSegment:
			if (!gaveBssError)
				uerror("opcode not allowed in BSS segment");
			gAddress[gSegment]++;
			gaveBssError = TRUE;
			break;
		default:
			assert(0);
	}
}

/*
 * EmitIvalue()
 */

void
EmitIvalue(IValue ivalue, int bigendian, int bytes)
{
	int		bit, value;

/* Get value/offset and make relocation entry if necessary */

	switch (ivalue->type)
	{
		case iSymOff:
			value = 0;
			RelocSym(&ivalue->u.symoff, bigendian, bytes);
			break;
		case iSegRel:
			value = ivalue->u.segrel.offset;
			if (ivalue->u.segrel.segment != AbsSegment)
				RelocSeg(&ivalue->u.segrel, bigendian, bytes);
			break;
		case iUnknown:
			value = 0;
			break;
		default:
			value = 0;
			assert(0);
	}

/* Emit bytes as specified endianness and bytes */

	if (bigendian)
	{
		for (bit = 0x08; bit; bit >>= 1, value <<= 8)
			if (bytes & bit)
				Emit((Byte) (value >> 24));
	}
	else
		for (bit = 1; bit <= 0x08; bit <<= 1, value >>= 8)
			if (bytes & bit)
				Emit((Byte) value);
}

/*
 * EmitImm()
 *
 * Emit an immediate mode instruction
 */

void
EmitImm(unsigned ins, IValue ivalue, int hibyte)
{
	Emit(GetOp(ins, mImmediate));
	EmitIvalue(ivalue, FALSE, hibyte ? 0x02 : 0x01);
}

/*
 * EmitIns()
 *
 * Emit an instruction.
 */

void
EmitIns(unsigned ins, IValue ivalue, int mode, int zmode)
{

/* Is this an implicit mode instruction? */

	if (mode == mImplied)
	{
		assert(ivalue == NULL);
		if (CheckOp(ins, mImplied) >= 0)
			Emit(GetOp(ins, mImplied));
		else
		{
			uerror("implied mode is invalid here");
			Emit(0);
		}
		return;
	}

/* Is this a branch instruction? */

	if (CheckOp(ins, mRelative) >= 0)
	{
		EmitRel(ins, ivalue);
		return;
	}

/* Try to do zero page */

	if (zmode && CheckOp(ins, zmode) >= 0)
		if (ivalue->type == iSegRel
				&& ivalue->u.segrel.segment == AbsSegment
				&& (ivalue->u.segrel.offset & ~0xFF) == 0)
		{
			Emit(GetOp(ins, zmode));
			Emit(ivalue->u.segrel.offset);
			return;
		}
		else if (ivalue->type == iSymOff
				&& IsZpImport(ivalue->u.symoff.id))
		{
			Emit(GetOp(ins, zmode));
			EmitIvalue(ivalue, FALSE, 0x01);
			return;
		}

/* Can we ONLY do zero page (but didn't)? */

	if (!mode)
	{
		uerror("zero-page address required");
		return;
	}

/* Generic three-byte instruction */

	Emit(GetOp(ins, mode));
	EmitIvalue(ivalue, FALSE, 0x03);
}

/*
 * EmitRel()
 *
 * Emit a relative branch instruction
 */

static	void
EmitRel(int ins, IValue ivalue)
{
	int		disp = 0;

/* Emit opcode */

	Emit(GetOp(ins, mRelative));

/* Compute displacement to target address */

	if (gPass == 2)
		if (ivalue->type != iSegRel || ivalue->u.segrel.segment != gSegment)
			uerror("illegal branch address");
		else
		{
			disp = ivalue->u.segrel.offset - (gAddress[gSegment] + 1);
			if (disp < -128 || disp > 127)
				uerror("branch out of range (%d bytes)", disp);
		}

/* Done */

	Emit(disp);
}

/*
 * EmitDump()
 */

unsigned
EmitSegDump(FILE *fp, int segment)
{
	if (fwrite(gSegBuffer[segment], 1,
		gAddress[segment], fp) != gAddress[segment])
	{
		perror("fwrite");
		cerror("can't write segment");
	}
	return(gAddress[segment]);
}

/*
 * EmitShowBytes()
 *
 * Assumes no trailing newline in "line".
 */

/*
 * ShowLineNum()
 */

static	void
ShowLineNum(FILE *fp)
{
	fprintf(fp, "%4d", gLineNumber);
	if (gLineNumber < 10000)
		fprintf(fp, " ");
}

static	void
ShowLoc(FILE *fp, int segment, unsigned address)
{
	fprintf(fp, "%c %04X ",
		segment == CodeSegment ? 'C' :
		segment == DataSegment ? 'D' :
		'B', address);
}

void
EmitShowBytes(FILE *fp)
{
	int				nbytes, extra, k;
	static int		pass = 0;
	static int		segment;
	static unsigned	address;

/* Make sure we know where we are */

	if (pass != gPass)
	{
		segment = INITIAL_SEGMENT;
		address = 0;
		pass = gPass;
	}

/* Show line number */

	ShowLineNum(fp);

/* Show location for label or if bytes were emitted */

	if (gLocLabel || address != gAddress[segment])
		ShowLoc(fp, segment, address);
	else
		fprintf(fp, "\t");

/* Compute number of total and expanded bytes to show */

	if (segment == BssSegment)
		nbytes = 0;
	else
		nbytes = gAddress[segment] - address;

	if (nbytes > 4)
	{
		nbytes = 4;
		extra = gExpand ? nbytes - 4 : 0;
	}
	else
		extra = 0;

/* Always show up to four emitted bytes */

	for (k = 0; k < nbytes; k++)
	{
		fprintf(fp, "%02X", gSegBuffer[segment][address++]);
		if (k < nbytes - 1)
			fprintf(fp, " ");
	}
	if (nbytes <= 1)
		fprintf(fp, "\t");
	fprintf(fp, "\t");

/* Now show the input line */

	fprintf(fp, "%s\n", gLineBuf);

/* Show remaining bytes if expand is on */

	if (gExpand && segment != BssSegment)
		while (extra)
		{
			ShowLineNum(fp);
			ShowLoc(fp, segment, address);
			nbytes = extra;
			nbytes = (nbytes <= 8) ? nbytes : 8;
			extra -= nbytes;
			for (k = 0; k < nbytes; k++)
			{
				fprintf(fp, "%02X", gSegBuffer[segment][address++]);
				if (k < nbytes - 1)
					fprintf(fp, " ");
			}
			fprintf(fp, "\n");
		}

	fflush(fp);

/* Remember where we are now for next time */

	segment = gSegment;
	address = gAddress[gSegment];
}

